(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.50 2014-02-26 00:12:42 fclement Exp $ *)

module Nat = Unatural;;
module Cmd_opt = Ucommand_option;;

let string_of_list s_of l =
  let len = Nat.to_int (Ulist.length l) in
  Ulist.fold_left_i
    (fun i accu x ->
      match i with
      | 0 -> accu ^ s_of x
      | n when n = len - 1 -> accu ^ ", and " ^ s_of x
      | _ -> accu ^ ", " ^ s_of x)
    "" l
;;

let o_qual = Upervasives.qualify "Options";;

let o_failwith = Upervasives.failwith_qual o_qual;;
let o_invalid_arg = Upervasives.invalid_arg_qual o_qual;;

open Segmentation_types;;

(* Codomain implementation.*)

type option_codomain_implementation =
  | Option_Tuple_Cod
  | Option_Vector_Cod
;;

let default_codomain_implementation = Option_Tuple_Cod;;

module Codomain_implementation_structure =
struct
  type t = option_codomain_implementation;;
  
  let name = "codomain implementation"
  and plural_name = "codomain implementations"
  ;;
  
  let default = default_codomain_implementation;;
  
  let of_string = function
    | "tuple_codomain" | "tuple" | "tc" | "t" -> Option_Tuple_Cod
    | "vector_codomain" | "vector" | "vc" | "v" -> Option_Vector_Cod
    | s ->
        o_failwith "codomain_implementation_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Option_Tuple_Cod -> "tuple_codomain"
    | Option_Vector_Cod -> "vector_codomain"
  ;;
  
  let all = [ Option_Tuple_Cod; Option_Vector_Cod; ];;
end
;;

module Codomain_implementation =
  Ustring_command_option.Make (Codomain_implementation_structure)
;;
let get_option_codomain_implementation = Codomain_implementation.get;;

(* Dimension.*)

type option_dimension = int;;

let default_dimension = 3;;

let { Uaref.get = get_option_dimension;
      set = set_option_dimension; } =
  Uaref.make_aref_single "dimension" default_dimension
;;

let option_dimension =
  {
   Cmd_opt.key = "-dimension";
   spec = Arg.Int set_option_dimension;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet dimension of the image in each pixel to <nat>.\n\
        \tThe dimension is a positive natural number.\n\
        \tIt is 1 for monochrome images, and 3 for RGB images.\n\
        \tDefault dimension is %d."
       (get_option_dimension ());
  }
;;

(* Dimension management. *)

type option_dimension_management =
  | Option_Vector
  | Option_Best_component_only
  | Option_Best_component_for_each
  | Option_Combine_best_components
;;

let default_dimension_management = Option_Vector;;

module Dimension_management_structure =
struct
  type t = option_dimension_management;;
  
  let name = "dimension management"
  and plural_name = "dimension managements"
  ;;
  
  let default = default_dimension_management;;
  
  let of_string = function
    | "vector" | "v" -> Option_Vector
    | "best_component_only" | "best_only" | "bco" ->
        Option_Best_component_only
    | "best_component_for_each" | "each" | "bcfe" ->
        Option_Best_component_for_each
    | "combine_best_components" | "combine" | "cbc" ->
        Option_Combine_best_components
    | s ->
        o_failwith "dimension_management_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Option_Vector -> "vector"
    | Option_Best_component_only -> "best_component_only"
    | Option_Best_component_for_each -> "best_component_for_each"
    | Option_Combine_best_components -> "combine_best_components"
  ;;
  
  let all =
    [ Option_Vector; Option_Best_component_only;
      Option_Best_component_for_each; Option_Combine_best_components; ]
  ;;
end
;;

module Dimension_management =
  Ustring_command_option.Make (Dimension_management_structure)
;;
let get_option_dimension_management = Dimension_management.get;;

(* Cutting strategy. *)

type option_cutting_strategy =
  | Option_Best_in_a_family
  | Option_Overall_best
;;

let default_cutting_strategy = Option_Overall_best;;

module Cutting_strategy_structure =
struct
  type t = option_cutting_strategy;;
  
  let name = "cutting strategy"
  and plural_name = "cutting strategies"
  ;;
  
  let default = default_cutting_strategy;;
  
  let of_string = function
    | "best_in_a_family" | "bf" | "family" | "f" -> Option_Best_in_a_family
    | "overall_best" | "ob" | "best" | "b" -> Option_Overall_best
    | s ->
        o_failwith "cutting_strategy_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Option_Best_in_a_family -> "best_in_a_family"
    | Option_Overall_best -> "overall_best"
  ;;
  
  let all = [ Option_Best_in_a_family; Option_Overall_best; ];;
end
;;

module type CUTTING_STRATEGY =
    Ustring_command_option_spec.T with type t = option_cutting_strategy
;;
module Cutting_strategy : CUTTING_STRATEGY =
  Ustring_command_option.Make (Cutting_strategy_structure)
;;
let get_option_cutting_strategy = Cutting_strategy.get;;

type option_number_of_cuts = int;;

let default_number_of_cuts = 0;;

let { Uaref.get = get_option_number_of_cuts;
      set = set_option_number_of_cuts; } =
  Uaref.make_aref_single "number of cuts" default_number_of_cuts
;;

let option_number_of_cuts =
  {
   Cmd_opt.key = "-number-of-cuts";
   spec = Arg.Int set_option_number_of_cuts;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet number of cuts for each basic cutting family to <nat>.\n\
        \tThe number of cuts is a natural number.\n\
        \tBy convention, 0 means all possible cuts of the family.\n\
        \tDefault number of cuts is %d."
       (get_option_number_of_cuts ());
  }
;;

type option_basic_cutting_family =
  | Option_Horizontal
  | Option_Vertical

and option_basic_cutting_families = option_basic_cutting_family list
;;

let all_basic_cutting_families = [ Option_Horizontal; Option_Vertical; ];;

let default_basic_cutting_families = all_basic_cutting_families;;

let { Uaref.get_list = get_option_basic_cutting_families;
      clear_list = clear_option_basic_cutting_families;
      add_item = add_option_basic_cutting_family; } =
  Uaref.make_aref_list default_basic_cutting_families
;;

let option_clear_basic_cutting_families =
  {
   Cmd_opt.key = "-clear-basic-cutting-families";
   spec = Arg.Unit clear_option_basic_cutting_families;
   doc = Printf.sprintf "\n\tClear the basic cutting family list.";
  }
;;

let option_basic_cutting_family_of_string = function
  | "horizontal" | "hor" | "h" -> Option_Horizontal
  | "vertical" | "ver" | "v" -> Option_Vertical
  | s ->
      o_failwith "basic_cutting_family_of_string"
        (Printf.sprintf "Unknown basic cutting family %S" s)
;;

let add_option_basic_cutting_family_from_string =
  Upervasives.compose
    add_option_basic_cutting_family
    option_basic_cutting_family_of_string
;;

let string_of_option_basic_cutting_family = function
  | Option_Horizontal -> "horizontal"
  | Option_Vertical -> "vertical"
;;

let print_all_basic_cutting_families () =
  string_of_list
    string_of_option_basic_cutting_family
    all_basic_cutting_families
;;

let print_default_basic_cutting_family_list () =
  string_of_list
    string_of_option_basic_cutting_family
    (get_option_basic_cutting_families ())
;;

let option_add_basic_cutting_family =
  {
   Cmd_opt.key = "-add-basic-cutting-family";
   spec = Arg.String add_option_basic_cutting_family_from_string;
   doc =
     Printf.sprintf
       "<str>\n\
        \tAdd <str> to the list of basic cutting families.\n\
        \tUsed when the cutting strategy is set to %S.\n\
        \tValid basic cutting families are: %t.\n\
        \tDefault list is %t."
       (Cutting_strategy_structure.to_string Option_Best_in_a_family)
       print_all_basic_cutting_families
       print_default_basic_cutting_family_list;
  }
;;

(* Stopping criterion. *)

type option_max_iteration = int;;

let default_max_iteration = 10;;

let { Uaref.get = get_option_max_iteration;
      set = set_option_max_iteration; } =
  Uaref.make_aref_single "max iteration" default_max_iteration
;;

let option_max_iteration =
  {
   Cmd_opt.key = "-max-iteration";
   spec = Arg.Int set_option_max_iteration;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet maximum number of iterations to <nat>.\n\
        \tIteration numbers are natural numbers.\n\
        \tDefault maximum number of iterations is %d."
       (get_option_max_iteration ());
  }
;;

type option_min_best_scalar_indicator = float;;

let default_min_best_scalar_indicator = 1e-2;;

let { Uaref.get = get_option_min_best_scalar_indicator;
      set = set_option_min_best_scalar_indicator; } =
  Uaref.make_aref_single
    "min best scalar indicator"
    default_min_best_scalar_indicator
;;

let option_min_best_scalar_indicator =
  {
   Cmd_opt.key = "-min-best-scalar-indicator";
   spec = Arg.Float set_option_min_best_scalar_indicator;
   doc =
     Printf.sprintf
       "<float>\n\
        \tSet minimum best scalar indicator to <float>.\n\
        \tIndicators are non negative real numbers.\n\
        \tDefault minimum best scalar indicator is %g."
       (get_option_min_best_scalar_indicator ());
  }
;;

type option_min_scalar_relative_error = float;;

let default_min_scalar_relative_error = 1e-2;;

let { Uaref.get = get_option_min_scalar_relative_error;
      set = set_option_min_scalar_relative_error; } =
  Uaref.make_aref_single
    "min scalar relative error"
    default_min_scalar_relative_error
;;

let option_min_scalar_relative_error =
  {
   Cmd_opt.key = "-min-scalar-relative-error";
   spec = Arg.Float set_option_min_scalar_relative_error;
   doc =
     Printf.sprintf
       "<float>\n\
        \tSet minimum scalar relative error to <float>.\n\
        \tRelative errors are non negative real numbers.\n\
        \tDefault minimum scalar relative error is %g."
       (get_option_min_scalar_relative_error ());
  }
;;

(* Program configuration. *)

type option_out_formatter =
  | Option_Std_out
  | Option_Std_err
  | Option_Out_file of Ufilename.t
;;

let default_option_out_formatter = Option_Std_out;;

module Out_formatter_structure =
struct
  type t = option_out_formatter;;
  
  let name = "out formatter"
  and plural_name = "out formatters"
  ;;
  
  let default = default_option_out_formatter;;
  
  let of_string = function
    | "-" | "stdout" | "Std_out" -> Option_Std_out
    | "stderr" | "Std_err" -> Option_Std_err
    | "" ->
        o_invalid_arg "out_formatter_of_string"
          "Empty out formatter specification"
    | s -> Option_Out_file s
  ;;
  
  let to_string = function
    | Option_Std_out -> "Std_out"
    | Option_Std_err -> "Std_err"
    | Option_Out_file fname -> fname
  ;;
  
  let all =
    [ Option_Std_out; Option_Std_err; Option_Out_file "<file name>" ; ]
  ;;
end
;;

module type OUT_FORMATTER =
    Ustring_command_option_spec.T with type t = option_out_formatter
;;
module Out_formatter : OUT_FORMATTER =
  Ustring_command_option.Make (Out_formatter_structure)
;;
let get_option_out_formatter = Out_formatter.get;;

type option_report_allocated_memory = bool;;

let { Uaref.get_flag = get_option_report_allocated_memory;
      set_flag = set_option_report_allocated_memory; _; } =
  Uaref.make_flag ()
;;

let option_report_allocated_memory =
  {
   Cmd_opt.key = Ustring.long_key_of_string "report allocated memory";
   spec = Arg.Unit set_option_report_allocated_memory;
   doc =
     Printf.sprintf "\n\tReport total allocated memory. Default is no report.";
  }
;;

type option_incremental_mode = bool;;

let { Uaref.get_flag = get_option_incremental_mode;
      set_flag = set_option_incremental_mode; _; } =
  Uaref.make_flag ()
;;

let option_incremental_mode =
  {
   Cmd_opt.key = Ustring.long_key_of_string "incremental mode";
   spec = Arg.Unit set_option_incremental_mode;
   doc =
     Printf.sprintf "\n\tTurn the incremental mode on. Default is off.";
  }
;;

type option_graphics_mode =
  | Option_No_graphics
  | Option_Minimal_graphics
  | Option_Parameter_graphics
  | Option_Compound_graphics
  | Option_Full_graphics
;;

let default_option_graphics_mode = Option_Full_graphics;;

module Graphical_mode_structure =
struct
  type t = option_graphics_mode;;
  
  let name = "graphics mode"
  and plural_name = "graphics modes"
  ;;
  
  let default = default_option_graphics_mode;;
  
  let of_string = function
    | "no_graphics" | "ng" -> Option_No_graphics
    | "minimal_graphics" | "mg" -> Option_Minimal_graphics
    | "parameter_graphics" | "pg" -> Option_Parameter_graphics
    | "compound_graphics" | "cg" -> Option_Compound_graphics
    | "full_graphics" | "fg" -> Option_Full_graphics
    | s ->
        o_failwith "graphics_mode_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;
  
  let to_string = function
    | Option_No_graphics -> "no_graphics"
    | Option_Minimal_graphics -> "minimal_graphics"
    | Option_Parameter_graphics -> "parameter_graphics"
    | Option_Compound_graphics -> "compound_graphics"
    | Option_Full_graphics -> "full_graphics"
  ;;
  
  let all =
    [ Option_No_graphics;
      Option_Minimal_graphics; Option_Parameter_graphics;
      Option_Compound_graphics ; Option_Full_graphics; ]
  ;;
end
;;

module type GRAPHICAL_MODE =
    Ustring_command_option_spec.T with type t = option_graphics_mode
;;
module Graphical_mode : GRAPHICAL_MODE =
  Ustring_command_option.Make (Graphical_mode_structure)
;;
let get_option_graphics_mode = Graphical_mode.get;;

type option_graphics_window_min_width = int
and option_graphics_window_min_height = int
and option_graphics_window_max_width = int
and option_graphics_window_max_height = int
;;

let default_graphics_window_min_width = 800
and default_graphics_window_min_height = 600
and default_graphics_window_max_width = 1920
and default_graphics_window_max_height = 1080
;;

let { Uaref.get = get_option_graphics_window_min_width;
      set = set_option_graphics_window_min_width; } =
  Uaref.make_aref_single
    "graphics window min width" default_graphics_window_min_width

and { Uaref.get = get_option_graphics_window_min_height;
      set = set_option_graphics_window_min_height; } =
  Uaref.make_aref_single
    "graphics window min height" default_graphics_window_min_height

and { Uaref.get = get_option_graphics_window_max_width;
      set = set_option_graphics_window_max_width; } =
  Uaref.make_aref_single
    "graphics window max width" default_graphics_window_max_width

and { Uaref.get = get_option_graphics_window_max_height;
      set = set_option_graphics_window_max_height; } =
  Uaref.make_aref_single
    "graphics window max height" default_graphics_window_max_height
;;

let option_graphics_window_min_width =
  {
   Cmd_opt.key = "-graphics-window-min-width";
   spec = Arg.Int set_option_graphics_window_min_width;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window minimum width to <nat>.\n\
        \tWidths are natural numbers.\n\
        \tDefault graphics window minimum width is %d."
       (get_option_graphics_window_min_width ());
  }

and option_graphics_window_min_height =
  {
   Cmd_opt.key = "-graphics-window-min-height";
   spec = Arg.Int set_option_graphics_window_min_height;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window minimum height to <nat>.\n\
        \tHeights are natural numbers.\n\
        \tDefault graphics window minimum height is %d."
       (get_option_graphics_window_min_height ());
  }

and option_graphics_window_max_width =
  {
   Cmd_opt.key = "-graphics-window-max-width";
   spec = Arg.Int set_option_graphics_window_max_width;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window maximum width to <nat>.\n\
        \tWidths are natural numbers.\n\
        \tDefault graphics window maximum width is %d."
       (get_option_graphics_window_max_width ());
  }

and option_graphics_window_max_height =
  {
   Cmd_opt.key = "-graphics-window-max-height";
   spec = Arg.Int set_option_graphics_window_max_height;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics window maximum height to <nat>.\n\
        \tHeights are natural numbers.\n\
        \tDefault graphics window maximum height is %d."
       (get_option_graphics_window_max_height ());
  }
;;

type option_graphics_gradient_threshold = int;;

let default_graphics_gradient_threshold = 10;;

let { Uaref.get = get_option_graphics_gradient_threshold;
      set = set_option_graphics_gradient_threshold; } =
  Uaref.make_aref_single
    "graphics gradient threshold" default_graphics_gradient_threshold
;;

let option_graphics_gradient_threshold =
  {
   Cmd_opt.key = "-graphics-gradient-threshold";
   spec = Arg.Int set_option_graphics_gradient_threshold;
   doc =
     Printf.sprintf
       "<nat>\n\
        \tSet graphics gradient threshold to <nat>.\n\
        \tThresholds are natural numbers.\n\
        \tDefault graphics gradient threshold is %d."
       (get_option_graphics_gradient_threshold ());
  }
;;

(* Initial and final images. *)

let init_initial_image_file_name = "";;

let { Uaref.set = set_argument_initial_image_file_name; _; } =
  Uaref.make_aref_unique
    "initial image file name" init_initial_image_file_name
;;

type argument_initial_image = image
and argument_initial_image_width = width
and argument_initial_image_height = height
;;

let init_initial_image = Umatrix.of_array_array [||]
and init_initial_image_width = Nat._0_
and init_initial_image_height = Nat._0_
;;

let { Uaref.get = get_argument_initial_image;
      set = set_argument_initial_image; } =
  Uaref.make_aref_unique "initial image" init_initial_image
and { Uaref.get = get_argument_initial_image_width;
      set = set_argument_initial_image_width; } =
  Uaref.make_aref_unique "initial image width" init_initial_image_width
and { Uaref.get = get_argument_initial_image_height;
      set = set_argument_initial_image_height; } =
  Uaref.make_aref_unique "initial image height" init_initial_image_height
;;

let set_initial_image fname =
  set_argument_initial_image_file_name fname;
  let pixmap = Ppm.load_pixmap fname in
  let idata = Umatrix.of_array_array pixmap in
  set_argument_initial_image idata;
  let (h, w) = Umatrix.size idata in
  if h = Nat._0_ || w = Nat._0_ then
    begin print_endline "empty image, nothing to do"; exit 0 end else
  begin set_argument_initial_image_width w;
  set_argument_initial_image_height h end;
;;

type option_final_image_file_name = Ufilename.t;;

let default_final_image_file_name = "final_image.ppm";;

let { Uaref.get = get_option_final_image_file_name;
      set = set_option_final_image_file_name; } =
  Uaref.make_aref_single "final image file name" default_final_image_file_name
;;

let option_final_image_file_name =
  {
   Cmd_opt.key = "-final-image-file-name";
   spec = Arg.String set_option_final_image_file_name;
   doc =
     Printf.sprintf
       "<fname>\n\
        \tSet final image file name to <fname>.\n\
        \tDefault is %S."
       (get_option_final_image_file_name ());
  }
;;

(* Report/debug flags and error messages. *)

let purpose = Udebug.User;;
let software_name = Ref_image_version.pph_id_card.Upph_id_card.project_name;;
let key_prefixes = Udebug.make_key_prefixes software_name;;

let printing_facilities =
  Udebug.make_printing_facilities purpose software_name key_prefixes
;;

(* Command-line options. *)

let regular_opt rco =
  Cmd_opt.with_auto_short_key
    (Cmd_opt.with_auto_gnu_key
       (Cmd_opt.of_regular rco))
;;

let user_add_all l = Cmd_opt.user_add_all (Ulist.map regular_opt l);;

let regular_opt_with_short_key (sk, rco) =
  Cmd_opt.with_short_key sk
    (Cmd_opt.with_auto_gnu_key
       (Cmd_opt.of_regular rco))
;;

let user_add_all_with_short_key l =
  Cmd_opt.user_add_all (Ulist.map regular_opt_with_short_key l)
;;

let phase_dimension = Cmd_opt.Phase "dimension";;

let add_options () =
  user_add_all [ Codomain_implementation.option; ];
  user_add_all_with_short_key [ ("-dim", option_dimension); ];
  user_add_all
    [
     Dimension_management.option;
     option_number_of_cuts;
     option_clear_basic_cutting_families;
     option_add_basic_cutting_family;
     Cutting_strategy.option;
     option_max_iteration;
     option_min_best_scalar_indicator;
     option_min_scalar_relative_error;
     Out_formatter.option;
     option_report_allocated_memory;
     option_incremental_mode;
     Graphical_mode.option;
    ];
  user_add_all_with_short_key
    [
     ("-min-w", option_graphics_window_min_width);
     ("-min-h", option_graphics_window_min_height);
     ("-max-w", option_graphics_window_max_width);
     ("-max-h", option_graphics_window_max_height);
    ];
  user_add_all
    [
     option_graphics_gradient_threshold;
     option_final_image_file_name;
    ];
  Cmd_opt.phase_add_all
    phase_dimension
    [ "-codomain-implementation"; "-dimension"; ];
;;
add_options ();;

Cmd_opt.set_anon_fun set_initial_image;;
Cmd_opt.set_usage_msg
  (Format.sprintf
     "@.%s -- Segmentation of images@.@.Options are:@."
     Sys.argv.(0))
;;
