(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: segmentation_types.mli,v 1.69 2014-01-21 17:17:23 fclement Exp $ *)

type component_number = Unatural.t;;

(** {6 Classical notion of images} *)

type length = Upervasives.length
and width = length
and height = length
and size = width * height
;;

type color = Graphics.color;;
type image = color Umatrix.t;;

(** {6 External representation of images}
 Extensional types for image segmentation. *)

type point = Unatural.t * Unatural.t;;

type 'a gimage = 'a Umatrix_graph.t;;
(** An ['a gimage] is the graph of the function that associates ['a] values to
 each [point] of the image. For instance, ['a] may be an RGB color triple. *)

(** {6 Internal representation of images}
 Structural types for image segmentation. *)

type 'a indicator = 'a;;

type 'a zone_cutting_candidate =
  {
   indicator : 'a indicator;
   cutting : Umatrix_nonempty_set.dichotomy;
  }
;;
(** An ['a cutting_candidate] is the association of a dichotomy of the
 support of a sub-part of the image and of the corresponding refinement
 indicator. *)

type 'a best_zone_cutting =
  | No_more_zone_cutting
  | Best_zone_cutting of 'a zone_cutting_candidate
;;
(** An ['a best_cutting] is isomorphic to an ['a cutting_candidate] option. *)

type 'a cost = 'a;;

type 'a zone_constants =
  {
   zc_param : 'a;
   zc_cost : 'a cost;
  }
;;

type 'a zc_gimage = 'a zone_constants gimage;;

type 'a zone =
  ('a zone_constants, 'a best_zone_cutting) Umatrix_structural_mapping.block
;;
(** An ['a zone] is a sub-map of an ['a zone_constants gimage] and the
 associated best cutting. *)

type zone_number = Umatrix_structural_mapping.block_number;;
(** Zones are numbered. *)

type bounds =
  {
   ll : point;
   ur : point;
  }
;;
(** Bounds of zones are represented by lower left and upper right corners
 of the smallest rectangle containing the support of the zone. *)

type ('a, 'b) pimage =
  ('a zone_constants, 'a best_zone_cutting, 'b) Umatrix_structural_mapping.t
;;
(** An ['a pimage] is a collection of ['a zone].
 The intention is to only consider zones whith disjoint supports.
 ['a] will always be scalar. *)

type ('a, 'b) cimage = 'a cost * ('a, 'b) pimage;;

type ('a, 'b) zonation =
  {
   component : component_number;
   other_components : component_number list;
   gimage_data : 'a gimage;
   cimage : ('a, 'b) cimage;
  }
;;
(** Zonations also carry the initial image to be segmented. *)

(** {6 Segmentation algorithm} *)

type output_status =
  | Max_iteration of History_types.state_number
  | Min_scalar_indicator of float
  | Min_scalar_relative_error of float
  | No_more_cuttings
;;

type number_of_cuts = Unatural.t;;

type basic_cutting_family =
  | Horizontal
  | Vertical
;;

type cutting_strategy =
  | Best_in_a_family of number_of_cuts * basic_cutting_family Ulist1.list1
  | Overall_best
;;

type 'a cutting =
  cutting_strategy -> 'a gimage -> 'a zc_gimage -> 'a best_zone_cutting
;;

type 'a component_cutting_candidate =
  {
   zone_number : zone_number;
   zone_cutting_candidate : 'a zone_cutting_candidate;
  }
;;

type 'a best_component_cutting =
  | No_more_component_cutting
  | Best_component_cutting of 'a component_cutting_candidate
;;

type 'a global_cutting_candidate =
  {
   component_number : component_number;
   component_cutting_candidate : 'a component_cutting_candidate;
  }
;;

type 'a best_global_cutting =
  | No_more_global_cutting
  | Best_global_cutting of 'a global_cutting_candidate
;;
