(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udelayed_dgraph.ml,v 1.10 2014/02/09 17:27:28 fclement Exp $ *)

(* Delayed dynamic graphs. *)

module Nat = Unatural;;
module Sdarray = Ustamped_darray;;
module Dgraph = Udynamic_graph;;

open Dgraph;;

let ddg_qual = Upervasives.qualify "Delayed_dgraph";;

let ddg_invalid_arg = Upervasives.invalid_arg_qual ddg_qual;;

type arrow_number = Dgraph.arrow_number;;

type 'b delayed_image =
  | Not_yet_computed
  | Image of 'b
;;

type ('a, 'b) delayed_arrow = ('a, 'b delayed_image) Dgraph.arrow;;

type ('a, 'b, 'c) t =
  {
   f : 'c -> arrow_number -> 'a -> 'b;
   graph : ('a, 'b delayed_image) Dgraph.t;
   to_do : arrow_number list;
  }
;;

let get ddg = Dgraph.get ddg.graph;;

let set ddg i x =
  Dgraph.set ddg.graph i { value = x; image = Not_yet_computed; };
  { ddg with to_do = i :: ddg.to_do; }
;;

let set_next ddg = set ddg (Dgraph.next_unset ddg.graph);;

let make f n x =
  if n = Unatural._0_ then ddg_invalid_arg "make" "empty delayed graph" else
  let ddg = { f = f; graph = Dgraph.make n; to_do = []; } in
  set ddg Unatural._0_ x
;;

let same ddg1 ddg2 =
  ddg1.f == ddg2.f &&
  Dgraph.same ddg1.graph ddg2.graph &&
  ddg1.to_do = ddg2.to_do
;;

let do_it param ({ f = f; graph = dg; to_do = l; } as ddg) =
  let do_it_f i =
    match Dgraph.get dg i with
    | ({ value = x; image = Not_yet_computed; } as ar) ->
        let ar = { ar with image = Image (f param i x); } in
        Dgraph.set dg i ar
    | { image = Image _; _; } -> assert false in
  Ulist.iter do_it_f l;
  { ddg with to_do = []; }
;;

let check_nonempty_to_do fun_name f = function
  | { to_do = _ :: _; _; } -> ddg_invalid_arg fun_name "nonempty to_do list"
  | ({ to_do = []; _; } as ddg) -> f ddg
;;

(* Converters. *)

let get_image = function
  | Not_yet_computed -> assert false
  | Image x -> x
;;

let get_arrow _an ({ image = dim; _; } as ar) =
  { ar with image = get_image dim; }
;;

let unsafe_to_array ddg = Dgraph.to_array (Dgraph.map get_arrow ddg.graph);;

let to_array ddg = check_nonempty_to_do "to_array" unsafe_to_array ddg;;

let delay _ = Not_yet_computed;;

let of_values f a =
  let dg = Dgraph.of_values delay a in
  let len = Dgraph.next_unset dg in
  let l = Ulist.of_range Nat.pred Nat._0_ len in
  { f = f; graph = dg; to_do = l; }
;;

(* Iterators. *)

let unsafe_fold_left { graph = dg; _; } f accu =
  let ff accu i = function
    | { image = Not_yet_computed; _ } -> assert false
    | ({ image = Image _; _ } as ar) -> f accu i ar in
  Dgraph.fold_left ff accu dg
;;

let fold_left f accu ddg =
  check_nonempty_to_do "fold_left" unsafe_fold_left ddg f accu
;;

let iter f ddg =
  check_nonempty_to_do "iter" unsafe_fold_left ddg (fun _ i ar -> f i ar) ()
;;

let iter_to_do f ddg =
  let ff i =
    let ar = get ddg i in
    f i ar.value in
  Ulist.iter ff ddg.to_do
;;

let iter_value f ddg =
  let ff i ar = f i ar.value in
  Dgraph.iter ff ddg.graph
;;

let fold_left_value f accu ddg =
  let ff accu i ar = f accu i ar.value in
  Dgraph.fold_left ff accu ddg.graph
;;
